/*
    SmartDongle libusb driver.

    MicroWorks, Inc.
    2808 North Cole Road
    Boise, ID 83704

    www.smartdongle.com


    ***** BEGIN LICENSE BLOCK *****

    Version: MPL 1.1/GPL 2.0/LGPL 2.1

    The contents of this file are subject to the Mozilla Public License
    Version 1.1 (the "License"); you may not use this file except in
    compliance with the License. You may obtain a copy of the License at
    http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS IS" basis,
    WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
    for the specific language governing rights and limitations under the
    License.

    The Original Code is MicroWorks, Inc. code.

    The Initial Developer of the Original Code is MicroWorks, Inc. Portions
    created by MicroWorks, Inc. are Copyright (C) 2008
    MicroWorks, Inc. All Rights Reserved.

    Contributor(s):

    Alternatively, the contents of this file may be used under the terms of
    either the GNU General Public License Version 2 or later (the "GPL"), or
    the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
    in which case the provisions of the GPL or the LGPL are applicable instead
    of those above. If you wish to allow use of your version of this file only
    under the terms of either the GPL or the LGPL, and not to allow others to
    use your version of this file under the terms of the MPL, indicate your
    decision by deleting the provisions above and replace them with the notice
    and other provisions required by the GPL or the LGPL. If you do not delete
    the provisions above, a recipient may use your version of this file under
    the terms of any one of the MPL, the GPL or the LGPL.

    ***** END LICENSE BLOCK *****

ChangeLog:
    v0.1.x  Initial driver code.

    v0.2.0  Support for multiple SmartDongles.

    v0.2.1  Cleaned up, some minor changes

    v0.3    usb_register_dev (>=2.5.x kernels)

    v0.4    Support for libusb user space driver, no synchronization
              is provide for libusb access.  v0.5    For SmartDongle firmware 1.04 and above

    v0.6    Read / write user EPROM space encryption mode via additional
              vendor USB commands.

    v0.6.1  Merged USB framework for 2.4.x kernels, for modules of course.
              Don't break out of wait loops (page wait, busy, twc) on polling
              IO errors - retry.

    v0.7.0  libusb 0.1.8 intermitienly reports errors sending control messages
              so retry USE_USBDEVFS_CMD_RETRIES.

    v0.8.0  Adding SmartDongle v1.5 functionality, Tested against FreeBSD 5.4
              NOTE: Will no longer support SmartDongle kernel modules.

    v0.8.1  ULL macro for all unsigned long long ints to fix compiler
              warning "integer constant is too large for 'long' type"

    v0.9.0  Moved some useful helper functions definitions to smartdongle.h
              Allow manual enable/disable outside of smrtdnglRead() and
              smrtdnglWrite().
              Modified smrtdnglSendControlMessage() to allow libusb errors
              from commands usb_open() and usb_claim_interface() to be
              returned to user. Error codes added for the above errors in
              smartdongle.h

    v0.10.0 Added GetErrorString(), GetEepromSize() and GetSerialNumber
            functions to match Windows API functionality as per customer
            request.

    v0.11.0 Check for endianness in order to support MAC arch. User function
            name have "SmartDongle" prefix to match corresponding function
            names of Windows API.  Internal functions are now static.

    v0.12.0 Overlooked ReadCryptMode in smartdongle.h in last version,
            renamed to SmartDongleReadCryptMode

    v0.12.1 Ported from SmartDongle libusb code version v0.12.0 for JNI

*/
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#if defined(SMARTDONGLE_JNI)
#include <jni.h>
#endif

#include "usb.h"
#include "smartdongle.h"

#if defined(SMARTDONGLE_JNI)
/* This is the header file generated with javah utility */
#include "SED_YYY.h"
#endif


/*
    SmartDongle state byte values
*/
#define SMARTDNGL_STATE0 0x00
#define SMARTDNGL_STATE1 0x01

#define SMARTDNGL_TWC0 0x00
#define SMARTDNGL_TWC1 0xFF

#define SMARTDNGL_BUSY0 0x00
#define SMARTDNGL_BUSY1 0xFF

/* SmartDongle v1.70 or above, extended functionality */
#define SMARTDNGL_OPTION_EXTENDED_SUPPORT2 0x20

/* SmartDongle v1.50 or above, extended functionality */
#define SMARTDNGL_OPTION_EXTENDED_SUPPORT 0x80

/* SmartDongle requires a new key for each access */
#define SMARTDNGL_OPTION_STRICT_ONLY      0x40

/* The firmware support bitmap */
#define SMARTDNGL_FIRMWARE_EXTENDED_SUPPORT2  0x40
#define SMARTDNGL_FIRMWARE_EXTENDED_SUPPORT  0x08
#define SMARTDNGL_FIRMWARE_STRICT_ONLY  0x04

#define SMARTDNGL_QTS_READ_RETRIES 3
#define SMARTDNGL_QTS_WRITE_RETRIES 3


/*
  Various timeouts (uSec)
*/
#define SMARTDNGL_PAGE_TIMEOUT   500000 /* EPROM page(s) ready */
#define SMARTDNGL_STATE0_TIMEOUT 250000 /* SmartDongle state reset */
#define SMARTDNGL_STATE1_TIMEOUT 250000 /* SmartDongle challenge ready */
#define SMARTDNGL_LOCK_TIMEOUT   250000
#define SMARTDNGL_UNLOCK_TIMEOUT 250000
#define SMARTDNGL_TWC_TIMEOUT    200000 /* SPI EPROM Time Write Cycle */
#define SMARTDNGL_BSY_TIMEOUT    500000 /* SmartDongle processing request */

/* Poll for SmartDongle state change */
#define SMARTDNGL_POLL_RETRYS       100

/* Poll for SmartDongle EEPROM prefetch ready */
#define SMARTDNGL_PAGE_TIMEOUT_POLL 50

/* For usb_control_msg timeout */
#define SMARTDNGL_MESG_TIMEOUT      1000
#define USE_USBDEVFS_CMD_RETRIES    3

/* USB Low speed packet size */
#define SMARTDNGL_PACKETSIZE        8

/*
    USB OUT SmartDongle Vendor Specific Requests
*/
#define SMARTDNGL_SEND_REPLY              0x00
#define SMARTDNGL_SEND_LED                0x01
#define SMARTDNGL_SEND_INFO               0x02
#define SMARTDNGL_SEND_CHALLENGE_REQ      0x03
#define SMARTDNGL_SEND_STATE_RESET        0x04
#define SMARTDNGL_SEND_SECURITY_RESET     0x05
#define SMARTDNGL_SEND_EPROM_PAGE         0x06
#define SMARTDNGL_SEND_EPROM_PAGE_REQ     0x07
#define SMARTDNGL_SEND_SOFT_RESET         0x09
#define SMARTDNGL_SEND_SET_EPROM_DEFAULT  0x0B
#define SMARTDNGL_SEND_SET_MODE           0x0C
#define SMARTDNGL_SEND_SET_STRICT_ONLY    0x0D

/*
    USB IN SmartDongle Vendor Specific Requests
*/
#define SMARTDNGL_GET_CHALLENGE           0x00
#define SMARTDNGL_GET_SECURITY_STATE      0x02
#define SMARTDNGL_GET_SD_STATE            0x03
#define SMARTDNGL_GET_EPROM_PAGE          0x04
#define SMARTDNGL_GET_EPROM_PAGE_ERROR    0x05
#define SMARTDNGL_GET_TWC_STATE           0x06
#define SMARTDNGL_GET_BUSY_STATE          0x07
#define SMARTDNGL_GET_OPTIONS             0x08
#define SMARTDNGL_GET_MODE                0x0A

/*
    Page flag bits
*/
#define SMARTDNGL_RDPAGEVALID          0x01
#define SMARTDNGL_WRPAGEVALID          0x02

/*
    EPROM read page error bits
*/
#define SMARTDNGL_RDADDRESS            0x04
#define SMARTDNGL_RDDENY               0x08
#define SMARTDNGL_RDEPROMERROR         0x10

/* EPROM write page error bits */
#define SMARTDNGL_WRADDRESS            0x20
#define SMARTDNGL_WRDENY               0x40
#define SMARTDNGL_WREPROMERROR         0x80

/* EPROM data error bits */
#define SMARTDNGL_RDTIMEOUT            0x01
#define SMARTDNGL_RDDATA               0x02
#define SMARTDNGL_WRTIMEOUT            0x10
#define SMARTDNGL_WRDATA               0x20

/* LED Port */
#define SMARTDNGL_LEDPORT              0x00

/*
    Descriptor types
*/
#define USB_DT_DEVICE    0x01
#define USB_DT_CONFIG    0x02
#define USB_DT_STRING    0x03
#define USB_DT_INTERFACE 0x04
#define USB_DT_ENDPOINT  0x05

#define USB_DT_HUB       0x29
#define USB_DT_HID       0x21
#define USB_DT_REPORT    0x22
#define USB_DT_PHYSICAL  0x23

/*
    Standard requests
*/
#define USB_REQ_GET_STATUS        0x00
#define USB_REQ_CLEAR_FEATURE     0x01
/* 0x02 is reserved */
#define USB_REQ_SET_FEATURE       0x03
/* 0x04 is reserved */
#define USB_REQ_SET_ADDRESS       0x05
#define USB_REQ_GET_DESCRIPTOR    0x06
#define USB_REQ_SET_DESCRIPTOR    0x07
#define USB_REQ_GET_CONFIGURATION 0x08
#define USB_REQ_SET_CONFIGURATION 0x09
#define USB_REQ_GET_INTERFACE     0x0A
#define USB_REQ_SET_INTERFACE     0x0B
#define USB_REQ_SYNCH_FRAME       0x0C

/*
    USB Request Type and Endpoint Directions
*/
#define USB_DIR_OUT                0
#define USB_DIR_IN                 0x80

#define USB_ENDPOINT_NUMBER_MASK   0x0f /* in bEndpointAddress */
#define USB_ENDPOINT_DIR_MASK      0x80

#define USB_ENDPOINT_XFERTYPE_MASK 0x03 /* in bmAttributes */
#define USB_ENDPOINT_XFER_CONTROL  0
#define USB_ENDPOINT_XFER_ISOC     1
#define USB_ENDPOINT_XFER_BULK     2
#define USB_ENDPOINT_XFER_INT      3

/*
    HIDD requests
*/
#define USB_REQ_GET_REPORT   0x01
#define USB_REQ_GET_IDLE     0x02
#define USB_REQ_GET_PROTOCOL 0x03
#define USB_REQ_SET_REPORT   0x09
#define USB_REQ_SET_IDLE     0x0A
#define USB_REQ_SET_PROTOCOL 0x0B

#define USB_TYPE_STANDARD    (0x00 << 5)
#define USB_TYPE_CLASS       (0x01 << 5)
#define USB_TYPE_VENDOR      (0x02 << 5)
#define USB_TYPE_RESERVED    (0x03 << 5)

#define USB_RECIP_DEVICE     0x00
#define USB_RECIP_INTERFACE  0x01
#define USB_RECIP_ENDPOINT   0x02
#define USB_RECIP_OTHER      0x03

#define USB_HID_RPT_INPUT    0x01
#define USB_HID_RPT_OUTPUT   0x02
#define USB_HID_RPT_FEATURE  0x03

/*
    SmartDongle USB Vendor and Product ID
*/
#define SMARTDNGL_VENDOR_ID  0x0DBF
#define SMARTDNGL_PRODUCT_ID 0x0002

/*
    SmartDongle access state byte values
*/
#define SMARTDNGL_LOCKED   0x00
#define SMARTDNGL_UNLOCKED 0x01

/* Challenge/Reply size, in bytes */
#define SMARTDNGL_KEYSIZE 8

/*
    SmartDongle EEPROM parameters
*/
#define SMARTDNGL_EPROMPGMASK   0xFFF0 /* Last nibble is "don't care" */
#define SMARTDNGL_EPROMPREFETCH 0x08
#define SMARTDNGL_MAXPREFETCH   (SMARTDNGL_EPROMPGSZ*SMARTDNGL_EPROMPREFETCH)
#define SMARTDNGL_PREFETCH_SIZE SMARTDNGL_MAXPREFETCH

/*
    usb_init(), usb_find_busses() have been called in main()
*/
#define SMARTDNGL_LIBUSB_PREINIT 0x02

/* SmartDongle device instance io buffer size */
#define SMARTDNGL_IOBUFSZ 1024

/* LED control states */
#define SMARTDNGL_RED                  0xFD
#define SMARTDNGL_GREEN                0xFE
#define SMARTDNGL_OFF                  0xFF

#define USK_ERR_JNI 254

/*
    SmartDongle control structure
*/

typedef struct smartdngl_control_t {
    int error;
    unsigned char status;
} smartdngl_control;

/* SmartDongle device instance data */
typedef struct smartdngl_dev_t {
    struct usb_device *dev;
    char iobuffer[SMARTDNGL_IOBUFSZ];
    unsigned char opt;

    // Record for system endianness
    int big_e;

    // These vars record for byte order
    unsigned short byte0;
    unsigned short byte1;
    unsigned short sig;
    unsigned short sig2;
    int index;
    usb_dev_handle *hdev;
} smartdngl_dev;

static smartdngl_dev Dongle;

#if defined(SMARTDONGLE_JNI)
jint lastError = 0;
#endif

/*
    P1      Pointer to key #1
    P2      Pointer to key #2
    address Address into SmartDongle user memory. (0 thru 32431)
    buffer  Pointer to user allocated data buffer.
    size    Requested number of bytes to read from SmartDongle.
    dongle  Pointer to dongle instance.
    control Pointer to control structure.
*/
static void smrtdnglRead(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned short address,
    unsigned char *buffer,
    int size,
    smartdngl_dev *dongle,
    smartdngl_control *control);

/*
 address Address into SmartDongle user memory. (0 thru 32431)
 buffer  Pointer to user allocated data buffer.
 size    Requested number of bytes to read from SmartDongle.
 dongle  Pointer to dongle instance.
 control Pointer to control structure.
 */
static void smrtdnglKeylessRead(
	 unsigned short address,
	 unsigned char *buffer,
	 int size,
	 smartdngl_dev *dongle,
	 smartdngl_control *control);

/*
    P1      Pointer to key #1
    P2      Pointer to key #2
    address Address into SmartDongle user memory. (0 thru 32431)
    buffer  Pointer to user allocated data buffer.
    size    Requested number of bytes to write to SmartDongle.
    dongle  Pointer to dongle instance.
    control Pointer to control structure.
*/
static void smrtdnglWrite(
    unsigned long long *P1,
	unsigned long long *P2,
	unsigned short address,
    unsigned char *buffer,
    int size,
    smartdngl_dev *dongle,
    smartdngl_control *control);

/*
 address Address into SmartDongle user memory. (0 thru 32431)
 buffer  Pointer to user allocated data buffer.
 size    Requested number of bytes to write to SmartDongle.
 dongle  Pointer to dongle instance.
 control Pointer to control structure.
 */
static void smrtdnglKeylessWrite(
	unsigned short address,
	unsigned char *buffer,
	int size,
	smartdngl_dev *dongle,
	smartdngl_control *control);

static void smrtdnglEnable(
    unsigned long long *P1,
    unsigned long long *P2,
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglValidateUsk(
    unsigned long long *P1,
    unsigned long long *P2,
    smartdngl_dev *dongle,
    smartdngl_control *control);

/*
    OUT Vendor Device Commands.
*/
static void smrtdnglSendReply(
    smartdngl_dev *dongle,
    smartdngl_control *control,
    unsigned long long *X);

static void smrtdnglSendChallenge(
    smartdngl_dev *dongle,
    smartdngl_control *control,
    unsigned long long *X);

static void smrtdnglLedControl(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglSendChallengeRequest(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglResetSdState(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglSendSecureStateReset(
    smartdngl_dev *dongle,
    smartdngl_control *control);
static void smrtdnglSendEepromPage(
    unsigned short address,
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglRequestEepromPrefetch(
    unsigned short address,
    smartdngl_dev *dongle,
    smartdngl_control *control);


#if !defined(SMARTDONGLE_JNI)
static void smrtdnglSendSoftReset(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglSendEpromMode(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglSendSetStrictOnly(
    smartdngl_dev *dongle,
    smartdngl_control *control);
#endif


/*
    IN Vendor Device Commands.
*/
static void smrtdnglGetKey(
    smartdngl_dev  *dongle,
    smartdngl_control *control,
    unsigned long long *X);

static void smrtdnglReadSecurityState(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglReadSdState(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglGetEpromPage(
    int *size,
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglReadPageFlags(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglReadTwcStatus(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglReadBusyStatus(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglGetOpt(
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglReadCryptMode(
    smartdngl_dev *dongle,
    smartdngl_control *control);

/*
    Wait Functions.
*/
static void smrtdnglWaitForPageRead(
    int timeout,
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglWaitForBusyStatus(
    int timeout,
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglWaitForSdState(
    int timeout,
    unsigned char state,
    smartdngl_dev *dongle,
    smartdngl_control *control);

static void smrtdnglWaitForTwcFlag(
    int timeout,
    smartdngl_dev *dongle,
    smartdngl_control *control);


static int smrtdnglProbe(smartdngl_control *control);

static smartdngl_dev* smrtdnglGetDev(smartdngl_control *control);

static int smrtdnglSendControlMessage(
    smartdngl_dev *dongle,
    unsigned char request,
    unsigned char requesttype,
    unsigned short value,
    unsigned short index,
    void *data,
    int size,
    int timeout);

static int smrtdnglGetDescriptor(
    smartdngl_dev *dongle,
    unsigned char type,
    unsigned char index,
    void *buf,
    int size);

static void intFlip(
   unsigned char *c);

static void endianConvert(
   unsigned char *data,
   unsigned int size);

/*
    sigFlip flips a short to dongle endianness
*/
static unsigned short sigFlip(unsigned short sig, smartdngl_dev *dongle);

#if defined(USB_STRING_DESCRIPTOR)
#undef USB_STRING_DESCRIPTOR
#endif

typedef struct USB_STRING_DESCRIPTOR {
    unsigned char  bLength ;
    unsigned char  bDescriptorType ;
    unsigned char  bString[1];  // pretending it's a simple string
} USB_STRING_DESCRIPTOR;

static int getSerialNumber(char *serNo, smartdngl_dev *smrtDongle, smartdngl_control *dnglControl);
static int getLoopCount(char *serialNumber);

#if !defined(SMARTDONGLE_JNI)
/*
    Given the correct keys, P1 and P2, will place the dongle into
    the enabled state and read EEPROM into buffer.
*/
int SmartDongleRead(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned short address,
    unsigned char *buffer,
    int size)
{
    smartdngl_control control;
    smartdngl_dev* dongle;

    dongle = smrtdnglGetDev(&control);
    if (dongle == NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle == NULL)
        {
            return control.error;
        }
    }

    /*
        Attempt to read user EEPROM space.
    */
    smrtdnglRead(P1, P2, address, buffer, size, dongle, &control);

    return control.error;
}

int SmartDongleKeylessRead(
	unsigned short address,
	unsigned char *buffer,
	int size)
{
    smartdngl_control control;
    smartdngl_dev* dongle;
	
    dongle = smrtdnglGetDev(&control);
    if (dongle == NULL) {
		
        control.status = 0;
        smrtdnglProbe(&control);
		
        dongle = smrtdnglGetDev(&control);
        if (dongle == NULL)
        {
            return control.error;
        }
    }
	
    /*
	 Attempt to read user EEPROM space.
	 */
    smrtdnglKeylessRead(address, buffer, size, dongle, &control);
	
    return control.error;
}

/*
    Given the correct keys, P1 and P2, will place the dongle into
    the enabled state and write buffer into EEPROM.
*/
int SmartDongleWrite(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned short address,
    unsigned char *buffer,
    int size)
{
    smartdngl_control control;
    smartdngl_dev *dongle;


    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }

    /*
        Attempt to write to user EEPROM space.
        Number of bytes written returned in size parameter.
    */
    smrtdnglWrite(P1, P2, address, buffer, size, dongle, &control);

    return control.error;
}

int SmartDongleKeylessWrite(
	unsigned short address,
	unsigned char *buffer,
	int size)
{
    smartdngl_control control;
    smartdngl_dev *dongle;
	
	
    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {
		
        control.status = 0;
        smrtdnglProbe(&control);
		
        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }
	
    /*
	 Attempt to write to user EEPROM space.
	 Number of bytes written returned in size parameter.
	 */
    smrtdnglKeylessWrite(address, buffer, size, dongle, &control);
	
    return control.error;
}

/*
    Switch SmartDongle LED indicator to Green.
*/
int SmartDongleLedGreen(void)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }

    control.status = SMARTDNGL_GREEN;
    smrtdnglLedControl(dongle, &control);

    return control.error;
}

/*
    Switch SmartDongle LED indicator to Red.
*/
int SmartDongleLedRed(void)
{
    smartdngl_control control;
    smartdngl_dev* dongle;
    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {
        control.status = 0;
        smrtdnglProbe(&control);
        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL) {
            return control.error;
        }
    }
    control.status = SMARTDNGL_RED;
    smrtdnglLedControl(dongle, &control);
    return control.error;
}

/*
    Switch SmartDongle LED indicator to Off.
*/
int SmartDongleLedOff(void)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }

    control.status = SMARTDNGL_OFF;
    smrtdnglLedControl(dongle, &control);

    return control.error;
}

/*
    SmartDongle software reset
*/
int SmartDongleReset(void)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }

    smrtdnglSendSoftReset(dongle, &control);
    return control.error;
}

/*
    For SmartDongle version >=1.70

    Set the current user memory space encryption mode.

    Warning: If data not written on 16 byte page boundaries, an encryption mode
    change can corrupt your data.

    Version 1.50 requires a SmartDongle reset before mode can be set,
    this would require a re-scan of the USB bus since the device handle
    would be invalid upon reset.
*/
int SmartDongleSetCryptMode(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned char *mode)
{
    smartdngl_control control;
    smartdngl_dev* dongle;
    int status;


    if (mode == NULL) {
    return SMARTDNGL_ERR_BAD_ARG;
    }

    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }

    /* SmartDongle must be version 1.70 or higher */
    if ((dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT2) == 0) {

        /* Lock SmartDongle. */
        smrtdnglSendSecureStateReset(dongle, &control);

        return SMARTDNGL_ERR_REQUIRES_1_70;
    }

    smrtdnglEnable(P1, P2, dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        return control.error;
    }

    /* Update encryption mode */
    control.status = *mode;
    smrtdnglSendEpromMode(dongle, &control);

    /* Lock SmartDongle. Ignore error.*/
    status = control.error;
    smrtdnglSendSecureStateReset(dongle, &control);

    return status;
}

/*
    Returns SmartDongle EPROM size
*/
int SmartDongleGetEepromSize (
    unsigned long long *P1,
    unsigned long long *P2)
{
    int status;
    unsigned char buffer[8]; // dummy buffer


    /*
        SmartDongle has 32K EPROM?
    */
    status = SmartDongleRead(P1, P2, 0x7EAF, buffer, 8);
    if (status == SMARTDNGL_ERR_OK)
    {
        return 0x7EB0;
    }

    /*
        SmartDongle has 8K EPROM?
    */
    status = SmartDongleRead(P1, P2, 0x1EAF, buffer, 8);
    if (status == SMARTDNGL_ERR_OK)
    {
        return  0x1EB0;
    }

    /*
        Unknown size or read error.
    */
    return -status;
}

/*****************************************************/
/*****************************************************/



/*
    For SmartDongle version 1.42 and higher.

    Set the current user memory space encryption mode.
*/
int SmartDongleReadCryptMode(unsigned char *mode)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    if (mode == NULL) {
    return SMARTDNGL_ERR_BAD_ARG;
    }

    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }

    smrtdnglReadCryptMode(dongle, &control);
    *mode = control.status;
    return control.error;
}

/*
    For SmartDongle >=v1.50 using >= v2.3 driver only.
    This will configure the SmartDongle for Strict Access.
    A new key and 32 bit signatures are required for each user
    read/write transaction against user memory.

    Caution: This configuration is permanent.
                SmartDongle must be sent back to the manufacturer to
                be reprogrammed to use the older drivers (<= v2.2.0).
*/
int SmartDongleSetStrictOnly(void)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        control.status = 0;
        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }

    smrtdnglSendSetStrictOnly(dongle, &control);
    return control.error;
}
#endif

/*
 Get SmartDongle serial number.
 User must allocate a serNo buffer of at least SMARTDNGL_SN_LENGTH + 1
 */
int SmartDongleGetSerialNumber (char *serNo)
{
    int received;
    smartdngl_control control;
    smartdngl_dev* dongle;
    unsigned char descBuf[128];
    USB_STRING_DESCRIPTOR *usbDesc;
    int i;
	
	
    if (serNo == NULL) {
        return SMARTDNGL_ERR_BAD_ARG;
    }
	
    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {
		
        control.status = 0;
        smrtdnglProbe(&control);
		
        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            return control.error;
        }
    }
	
    /*
	 Wait for task_loop processing to finish
	 */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        return control.error;
    }
	
    received = smrtdnglGetDescriptor(
									 dongle,
									 USB_DT_STRING,
									 SMARTDNGL_SN_INDEX,
									 descBuf,
									 128);
	
    /*
	 Allow libusb errors from commands usb_open() and
	 usb_claim_interface() to be returned to user.
	 */
    if (received < (SMARTDNGL_SN_LENGTH + 2)) {
        if (received == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            return SMARTDNGL_ERR_GET_SN;
        }
        else {
            return -received;
        }
    }
	
    usbDesc = (USB_STRING_DESCRIPTOR *)descBuf;
	
    /* Returns a Unicode string and 2 UCHARS */
    if ((usbDesc->bLength <((SMARTDNGL_SN_LENGTH * 2) + 2)) ||
        (usbDesc->bLength > SMARTDNGL_SN_MAXLEN + 2))
    {
        return SMARTDNGL_ERR_DATA_FORMAT;
    }
	
    /* An Unicode to ASCII conversion sufficient for our purpose */
    for (i = 0; i< SMARTDNGL_SN_LENGTH; i++)
    {
        serNo[i] = usbDesc->bString[i*2];
    }
	
    serNo[i] = '\0';
	
    return SMARTDNGL_ERR_OK;
}

static int getSerialNumber(char *serNo, smartdngl_dev* smrtDongle, smartdngl_control* dnglControl)
{
    int received;
    unsigned char descBuf[128];
    USB_STRING_DESCRIPTOR *usbDesc;
    int i;
	
	
    if (serNo == NULL) {
		return SMARTDNGL_ERR_BAD_ARG;
    }
	
    /*
	 Wait for task_loop processing to finish
	 */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, smrtDongle, dnglControl);
    if (dnglControl->error != SMARTDNGL_ERR_OK) {
        return dnglControl->error;
    }
	
    received = smrtdnglGetDescriptor(
									 smrtDongle,
									 USB_DT_STRING,
									 SMARTDNGL_SN_INDEX,
									 descBuf,
									 128);
	
    /*
	 Allow libusb errors from commands usb_open() and
	 usb_claim_interface() to be returned to user.
	 */
    if (received < (SMARTDNGL_SN_LENGTH + 2)) {
        if (received == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            return SMARTDNGL_ERR_GET_SN;
        }
        else {
            return -received;
        }
    }
	
    usbDesc = (USB_STRING_DESCRIPTOR *)descBuf;
	
    /* Returns a Unicode string and 2 UCHARS */
    if ((usbDesc->bLength <((SMARTDNGL_SN_LENGTH * 2) + 2)) ||
        (usbDesc->bLength > SMARTDNGL_SN_MAXLEN + 2))
    {
        return SMARTDNGL_ERR_DATA_FORMAT;
    }
	
    /* An Unicode to ASCII conversion sufficient for our purpose */
    for (i = 0; i< SMARTDNGL_SN_LENGTH; i++)
    {
        serNo[i] = usbDesc->bString[i*2];
    }
	
    serNo[i] = '\0';
	
    return SMARTDNGL_ERR_OK;
}

int getLoopCount(char* serialNumber) {
    int numLoops = 1;
    unsigned char numLoopID = serialNumber[0];
    switch(numLoopID){
        case '0':
            numLoops = 1;
            break;
        case '1':
            numLoops = 3;
            break;
        case '2':
            numLoops = 5;
            break;
        case '3':
            numLoops = 7;
            break;
        case '4':
            numLoops = 11;
            break;
        case '5':
            numLoops = 13;
            break;
        case '6':
            numLoops = 17;
            break;
        case '7':
            numLoops = 19;
            break;
        case '8':
            numLoops = 23;
            break;
        case '9':
            numLoops = 29;
            break;
        case 'A':
            numLoops = 31;
            break;
        case 'B':
            numLoops = 37;
            break;
        case 'C':
            numLoops = 41;
            break;
        case 'D':
            numLoops = 43;
            break;
        case 'E':
            numLoops = 47;
            break;
        case 'F':
            numLoops = 53;
            break;
        default:
            numLoops = 1;
            break;
    }
    return numLoops;
}

static int smrtdnglProbe(smartdngl_control *control)
{
    struct usb_device *dev;
    struct usb_bus *bus, *busses;
    unsigned char libUsbPreinit = control->status & SMARTDNGL_LIBUSB_PREINIT;
    int i;
    char *cp;


    control->error = SMARTDNGL_ERR_OK;

    Dongle.dev = NULL;

    /*
        Flag set if previously evoked.
    */
    if (!libUsbPreinit) {
        usb_init();
        usb_find_busses();
    }

    usb_find_devices();
    busses = usb_get_busses();

    for (bus = busses; bus; bus = bus->next) {

        for (dev = bus->devices; dev; dev = dev->next) {

            if ((dev->descriptor.idVendor == SMARTDNGL_VENDOR_ID) &&
                 (dev->descriptor.idProduct == SMARTDNGL_PRODUCT_ID)) {

                Dongle.dev = dev;

                /*Endiannes code contributed by CFMC */
                i = 0; cp = (char*)&i; cp[3] = 1;
                if (i == 1) {
                    Dongle.big_e = 1;
                    Dongle.byte0 = 7;
                    Dongle.byte1 = 6;
                } else {
                    Dongle.big_e = 0;
                    Dongle.byte0 = 0;
                    Dongle.byte1 = 1;
                }

                /*
                   Get the SmartDongle options.
                */
                smrtdnglGetOpt(&Dongle, control);
                if (control->error != SMARTDNGL_ERR_OK) {
                    /* Lowest common denominator */
                    Dongle.opt = 0;
                }

                return 1;
            }
        }
    }

    return 0;
}

/*
    Return the SmartDongle device instance, if any.
*/
static smartdngl_dev* smrtdnglGetDev(smartdngl_control *control)
{
    if (Dongle.dev != NULL) {
        control->error = SMARTDNGL_ERR_OK;
        return &Dongle;
    } else {
		control->status = 0;
        control->error = SMARTDNGL_ERR_NODEV;
        return NULL;
    }
}



/***********************************************************************/
/***********************************************************************/



/*
    Send a control message via Pipe 0. If successful, a non-negative number is
    returned that represents the number of bytes transferred.
*/
static int smrtdnglSendControlMessage(
    smartdngl_dev *dongle,
    unsigned char request,
    unsigned char requesttype,
    unsigned short value,
    unsigned short index,
    void *data,
    int size,
    int timeout)
{
    int transferred;
    int i;


    dongle->hdev = usb_open(dongle->dev);
    if (!dongle->hdev) {
        return -SMARTDNGL_ERR_USB_OPEN;
    }

    if (usb_claim_interface(dongle->hdev, 0) != 0) {
        usb_close(dongle->hdev);
        return -SMARTDNGL_ERR_USB_CLAIM_INTERFACE;
    }

    if (dongle->big_e && data != NULL) {
        if (!(request == SMARTDNGL_SEND_EPROM_PAGE &&
            !(requesttype & USB_DIR_IN))) {
             endianConvert((unsigned char *)data, size);
         }
    }

    /*
        libusb known to have intermitient errors sending control messages,
        so retry on failure.
    */
    for (i = 0; i< USE_USBDEVFS_CMD_RETRIES; i++) {
        transferred = usb_control_msg(
            dongle->hdev,
            (int)requesttype,
            (int)request,
            (int)value,
            (int)index,
            (char *)data,
            (int)size,
            timeout);

        if (transferred >= 0)
            break;
    }

    if (dongle->big_e && data != NULL) {
        if (!(request == SMARTDNGL_GET_EPROM_PAGE &&
            requesttype & USB_DIR_IN)) {

            endianConvert((unsigned char *)data, size);
        }
    }

    usb_release_interface(dongle->hdev, 0);
    usb_close(dongle->hdev);

    if (transferred < 0) {
        return -SMARTDNGL_ERR_USB_CONTROL_MSG;
    }

    return transferred;
}


/*
    Request descriptor by type and index.
    If successful, a non-negative number is returned that represents the number
    of bytes transferred.
*/
static int smrtdnglGetDescriptor(
    smartdngl_dev *dongle,
    unsigned char type,
    unsigned char index,
    void *buf,
    int size)
{
    int transferred;
    int i;

    dongle->hdev = usb_open(dongle->dev);
    if (!dongle->hdev) {
        return -SMARTDNGL_ERR_USB_OPEN;
    }

    if (usb_claim_interface(dongle->hdev, 0) != 0) {
        usb_close(dongle->hdev);
        return -SMARTDNGL_ERR_USB_CLAIM_INTERFACE;
    }

    /*
        libusb known to have intermitient errors sending control messages,
        so retry on failure.
    */
    for (i=0; i< USE_USBDEVFS_CMD_RETRIES; i++) {

        transferred = usb_get_descriptor(
            dongle->hdev,
            type,
            index,
            buf,
            size);

        if (transferred > 0)
            break;
    }

    usb_release_interface(dongle->hdev, 0);
    usb_close(dongle->hdev);

    if (transferred < 0) {
        return -SMARTDNGL_ERR_USB_CONTROL_MSG;
    }

    return transferred;
}



/***********************************************************************/
/***********************************************************************/



/*
    Read from SmartDongle EEPROM, see smartdongle.h for returned errors.
*/
static void smrtdnglRead(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned short address,
    unsigned char *buffer,
    int size,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned short offset;
    unsigned short page0;
    int prefetches;
    int prefetch;
    int savedStatus;
    int retries;
    int received;
    unsigned short addr;
    unsigned char readErrorMask;
    unsigned char *memory = NULL;


    control->error = SMARTDNGL_ERR_OK;
    readErrorMask = SMARTDNGL_RDADDRESS|SMARTDNGL_RDDENY|SMARTDNGL_RDEPROMERROR;

    /*
        If  size == 0 or NULL buffer just return enable status.
    */
    if (size == 0 || buffer == NULL) {
        smrtdnglEnable(P1, P2, dongle, control);
        goto FAIL;
    }

    /*
        User may request address that is inside a page.
        Get the offset from the first page.
    */
    page0 = address & ~SMARTDNGL_EPROMPGMASK;

    /*
        EPROM uses paged address: all bits below page boundaries a "don't care"
    */
    addr = address & SMARTDNGL_EPROMPGMASK;

    /*
        Calculate total number of read operations required
    */
    prefetches = (page0 + size) / SMARTDNGL_PREFETCH_SIZE;
    if (((page0 + size) % SMARTDNGL_PREFETCH_SIZE) > 0) {
        prefetches++;
    }

    memory = (unsigned char *)malloc(prefetches * SMARTDNGL_PREFETCH_SIZE);
    if (memory == NULL) {
        control->error = SMARTDNGL_ERR_MEM;
        goto FAIL;
    }

    for (prefetch = 0; prefetch < prefetches; prefetch++) {

        offset = addr + (prefetch * SMARTDNGL_PREFETCH_SIZE);

        /*
            Must obtain new key for each read transaction
            if SmartDongle configured for strict access.
        */
        if ((dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT)&&
             (dongle->opt & SMARTDNGL_OPTION_STRICT_ONLY)) {

            smrtdnglEnable(P1, P2, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }
        }
        /*
            Enable SmartDongle on first memory access if non-strict
        */
        else if (prefetch == 0) {

            smrtdnglEnable(P1, P2, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }
        }


        retries = 0;
        while (1) {

            retries++;
            if (retries == SMARTDNGL_QTS_READ_RETRIES) {
                control->error = SMARTDNGL_ERR_RD_EPROM;
                goto FAIL;
            }

            /*
                Request prefetch starting at address + offset
            */
            smrtdnglRequestEepromPrefetch(offset, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                continue;
            }

            /*
                Wait for prefetch to complete
            */
            smrtdnglWaitForPageRead(SMARTDNGL_PAGE_TIMEOUT, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                continue;
            }

            /*
                Don't retry on an address range error
            */
            control->status &= readErrorMask;
            if (control->status) {

                control->error = SMARTDNGL_ERR_RD_EPROM;

                if (control->status &
                    (SMARTDNGL_RDDENY|SMARTDNGL_RDEPROMERROR)) {
                    continue;
                }
                else {
                    goto FAIL;
                }
            }

            break;
        }

        /*
            Read current prefetch from SmartDongle.
        */
        smrtdnglGetEpromPage(&received, dongle, control);
        if (control->error != SMARTDNGL_ERR_OK) {
            goto FAIL;
        }

        /*
            Must be full prefetch size.
        */
        if (received < SMARTDNGL_PREFETCH_SIZE) {
            control->error = SMARTDNGL_ERR_RD_EPROM;
            goto FAIL;
        }

        memcpy(
            memory + (prefetch * SMARTDNGL_PREFETCH_SIZE),
            dongle->iobuffer,
            SMARTDNGL_PREFETCH_SIZE);
    }


    /*
        Copy from memory buffer from page 0 offset to user buffer
    */
    memcpy(buffer, memory + page0, size);

FAIL:
    if (memory != NULL) {
        free(memory);
    }

    savedStatus = control->error;

    /* Attempt to Lock SmartDongle. */
    smrtdnglSendSecureStateReset(dongle, control);

    control->error = savedStatus;
    return;
}

static void smrtdnglKeylessRead(
	unsigned short address,
	unsigned char *buffer,
	int size,
	smartdngl_dev *dongle,
	smartdngl_control *control)
{
    unsigned short offset;
    unsigned short page0;
    int prefetches;
    int prefetch;
    int savedStatus;
    int retries;
    int received;
    unsigned short addr;
    unsigned char readErrorMask;
    unsigned char *memory = NULL;
	
	
    control->error = SMARTDNGL_ERR_OK;
    readErrorMask = SMARTDNGL_RDADDRESS|SMARTDNGL_RDDENY|SMARTDNGL_RDEPROMERROR;
	
    /*
	 User may request address that is inside a page.
	 Get the offset from the first page.
	 */
    page0 = address & ~SMARTDNGL_EPROMPGMASK;
	
    /*
	 EPROM uses paged address: all bits below page boundaries a "don't care"
	 */
    addr = address & SMARTDNGL_EPROMPGMASK;
	
    /*
	 Calculate total number of read operations required
	 */
    prefetches = (page0 + size) / SMARTDNGL_PREFETCH_SIZE;
    if (((page0 + size) % SMARTDNGL_PREFETCH_SIZE) > 0) {
        prefetches++;
    }
	
    memory = (unsigned char *)malloc(prefetches * SMARTDNGL_PREFETCH_SIZE);
    if (memory == NULL) {
        control->error = SMARTDNGL_ERR_MEM;
        goto FAIL;
    }
	
    for (prefetch = 0; prefetch < prefetches; prefetch++) {
		
        offset = addr + (prefetch * SMARTDNGL_PREFETCH_SIZE);
		
        retries = 0;
        while (1) {
			
            retries++;
            if (retries == SMARTDNGL_QTS_READ_RETRIES) {
                control->error = SMARTDNGL_ERR_RD_EPROM;
                goto FAIL;
            }
			
            /*
			 Request prefetch starting at address + offset
			 */
            smrtdnglRequestEepromPrefetch(offset, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                continue;
            }
			
            /*
			 Wait for prefetch to complete
			 */
            smrtdnglWaitForPageRead(SMARTDNGL_PAGE_TIMEOUT, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                continue;
            }
			
            /*
			 Don't retry on an address range error
			 */
            control->status &= readErrorMask;
            if (control->status) {
				
                control->error = SMARTDNGL_ERR_RD_EPROM;
				
                if (control->status &
                    (SMARTDNGL_RDDENY|SMARTDNGL_RDEPROMERROR)) {
                    continue;
                }
                else {
                    goto FAIL;
                }
            }
			
            break;
        }
		
        /*
		 Read current prefetch from SmartDongle.
		 */
        smrtdnglGetEpromPage(&received, dongle, control);
        if (control->error != SMARTDNGL_ERR_OK) {
            goto FAIL;
        }
		
        /*
		 Must be full prefetch size.
		 */
        if (received < SMARTDNGL_PREFETCH_SIZE) {
            control->error = SMARTDNGL_ERR_RD_EPROM;
            goto FAIL;
        }
		
        memcpy(
			   memory + (prefetch * SMARTDNGL_PREFETCH_SIZE),
			   dongle->iobuffer,
			   SMARTDNGL_PREFETCH_SIZE);
    }
	
	
    /*
	 Copy from memory buffer from page 0 offset to user buffer
	 */
    memcpy(buffer, memory + page0, size);
	
FAIL:
    if (memory != NULL) {
        free(memory);
    }
	
    savedStatus = control->error;
	
    /* Attempt to Lock SmartDongle. */
    smrtdnglSendSecureStateReset(dongle, control);
	
    control->error = savedStatus;
    return;
}

/*
    Write to a SmartDongle EEPROM, see smartdongle.h for returned errors.
*/
static void smrtdnglWrite(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned short address,
    unsigned char *buffer,
    int size,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned short offset;
    unsigned short page0, remainder;
    unsigned short addr;
    int retries;
    int page, pages;
    int savedStatus;
    unsigned char writeErrorMask =
        SMARTDNGL_WRADDRESS|SMARTDNGL_WRDENY|SMARTDNGL_WREPROMERROR;
    unsigned char *memory = NULL;


    /*
        If  size == 0 or NULL buffer just return enable status.
    */
    if (size == 0 || buffer == NULL) {
        smrtdnglEnable(P1, P2, dongle, control);
        goto FAIL;
    }

    /*
        User may request address that is inside a page.
        Get the offset from the first page.
    */
    page0 = address & ~SMARTDNGL_EPROMPGMASK;

    /*
        EPROM uses paged address: all bits below page boundaries a "don't care"
    */
    addr = address & SMARTDNGL_EPROMPGMASK;

    /*
        Calculate total number of write pages with respect to offset into the
        first SmartDongle user memory page.
    */
    pages = (size + page0) / SMARTDNGL_EPROMPGSZ;
    remainder = (size + page0) % SMARTDNGL_EPROMPGSZ;
    if (remainder) {
        pages++;
    }

    memory = (unsigned char *)malloc(pages * SMARTDNGL_EPROMPGSZ);
    if (memory == NULL) {
        control->error = SMARTDNGL_ERR_MEM;
        goto FAIL;
    }

    /*
        Require read user memory data from page 0?
    */
    if (page0) {

        smrtdnglRead(
            P1,
            P2,
            addr,
            memory,
            SMARTDNGL_EPROMPGSZ,
            dongle, control);
        if (control->error != SMARTDNGL_ERR_OK) {
            goto FAIL;
        }
    }

    /*
        Require read user memory data from last page of write operation?
    */
    if ((pages > 1) && remainder) {

        smrtdnglRead(
            P1,
            P2,
            addr + ((pages - 1) * SMARTDNGL_EPROMPGSZ),
            memory + ((pages - 1) * SMARTDNGL_EPROMPGSZ),
            SMARTDNGL_EPROMPGSZ,
            dongle,
            control);
        if (control->error != SMARTDNGL_ERR_OK) {
            goto FAIL;
        }
    }

    /*
        Fill in the memory buffer with user data from offset.
    */
    memcpy(memory + page0, buffer, size);

    /*
        Write all pages out to SmartDongle user memory
    */
    for (page = 0; page < pages; page++) {

        offset = addr + (page * SMARTDNGL_EPROMPGSZ);

        /*
            Must obtain new key for each write transaction
            if SmartDongle configured for strict access.
        */
        if ((dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT) &&
             (dongle->opt & SMARTDNGL_OPTION_STRICT_ONLY)) {

            smrtdnglEnable(P1, P2, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                return;
            }
        }
        /*
            Enable SmartDongle on first memory access if non-strict
        */
        else if (page == 0) {

            smrtdnglEnable(P1, P2, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                return;
            }
        }

        for (retries = 0; retries < SMARTDNGL_QTS_WRITE_RETRIES; retries++) {
            /*
                Write the page
            */
            memcpy(
                dongle->iobuffer,
                memory + (page * SMARTDNGL_EPROMPGSZ),
                SMARTDNGL_EPROMPGSZ);

            smrtdnglSendEepromPage(offset, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }

            /*
                Wait for task_loop processing to finish
                control->error is reset in smrtdnglReadBusyStatus().
            */
            smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }

            /* Any errors writing page to EEPROM? */
            smrtdnglReadPageFlags(dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }

            /* Don't retry if an address range error */
            control->status &= writeErrorMask;
            if (control->status) {

                control->error = SMARTDNGL_ERR_WR_EPROM;

                if (control->status &
                    (SMARTDNGL_WRDENY|SMARTDNGL_WREPROMERROR)) {

                    /* Ignore errors here */
                    smrtdnglWaitForTwcFlag(
                        SMARTDNGL_TWC_TIMEOUT, dongle, control);

                    continue;
                } else {
                    break;
                }
            }

            /* Wait here for Twc (Time Write cycle) to reset */
            smrtdnglWaitForTwcFlag(SMARTDNGL_TWC_TIMEOUT, dongle, control);
            if (control->status != SMARTDNGL_ERR_OK) {
                break;
            }

            break;
        }

        /*
            Error from retry loop: break here and fail
        */
        if (control->status != SMARTDNGL_ERR_OK) {
            break;
        }
    }

FAIL:
    if (memory != NULL) {
        free(memory);
    }

    savedStatus = control->error;

    /* Attempt to Lock SmartDongle. */
    smrtdnglSendSecureStateReset(dongle, control);

    control->error = savedStatus;
    return;
}

static void smrtdnglKeylessWrite(
	unsigned short address,
	unsigned char *buffer,
	int size,
	smartdngl_dev *dongle,
	smartdngl_control *control)
{
    unsigned short offset;
    unsigned short page0, remainder;
    unsigned short addr;
    int retries;
    int page, pages;
    int savedStatus;
    unsigned char writeErrorMask =
	SMARTDNGL_WRADDRESS|SMARTDNGL_WRDENY|SMARTDNGL_WREPROMERROR;
    unsigned char *memory = NULL;
	
    /*
	 User may request address that is inside a page.
	 Get the offset from the first page.
	 */
    page0 = address & ~SMARTDNGL_EPROMPGMASK;
	
    /*
	 EPROM uses paged address: all bits below page boundaries a "don't care"
	 */
    addr = address & SMARTDNGL_EPROMPGMASK;
	
    /*
	 Calculate total number of write pages with respect to offset into the
	 first SmartDongle user memory page.
	 */
    pages = (size + page0) / SMARTDNGL_EPROMPGSZ;
    remainder = (size + page0) % SMARTDNGL_EPROMPGSZ;
    if (remainder) {
        pages++;
    }
	
    memory = (unsigned char *)malloc(pages * SMARTDNGL_EPROMPGSZ);
    if (memory == NULL) {
        control->error = SMARTDNGL_ERR_MEM;
        goto FAIL;
    }
	
    /*
	 Require read user memory data from page 0?
	 */
    if (page0) {
		
        smrtdnglKeylessRead(
					 addr,
					 memory,
					 SMARTDNGL_EPROMPGSZ,
					 dongle, control);
        if (control->error != SMARTDNGL_ERR_OK) {
            goto FAIL;
        }
    }
	
    /*
	 Require read user memory data from last page of write operation?
	 */
    if ((pages > 1) && remainder) {
		
        smrtdnglKeylessRead(
					 addr + ((pages - 1) * SMARTDNGL_EPROMPGSZ),
					 memory + ((pages - 1) * SMARTDNGL_EPROMPGSZ),
					 SMARTDNGL_EPROMPGSZ,
					 dongle,
					 control);
        if (control->error != SMARTDNGL_ERR_OK) {
            goto FAIL;
        }
    }
	
    /*
	 Fill in the memory buffer with user data from offset.
	 */
    memcpy(memory + page0, buffer, size);
	
    /*
	 Write all pages out to SmartDongle user memory
	 */
    for (page = 0; page < pages; page++) {
		
        offset = addr + (page * SMARTDNGL_EPROMPGSZ);
				
        for (retries = 0; retries < SMARTDNGL_QTS_WRITE_RETRIES; retries++) {
            /*
			 Write the page
			 */
            memcpy(
				   dongle->iobuffer,
				   memory + (page * SMARTDNGL_EPROMPGSZ),
				   SMARTDNGL_EPROMPGSZ);
			
            smrtdnglSendEepromPage(offset, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }
			
            /*
			 Wait for task_loop processing to finish
			 control->error is reset in smrtdnglReadBusyStatus().
			 */
            smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }
			
            /* Any errors writing page to EEPROM? */
            smrtdnglReadPageFlags(dongle, control);
            if (control->error != SMARTDNGL_ERR_OK) {
                break;
            }
			
            /* Don't retry if an address range error */
            control->status &= writeErrorMask;
            if (control->status) {
				
                control->error = SMARTDNGL_ERR_WR_EPROM;
				
                if (control->status &
                    (SMARTDNGL_WRDENY|SMARTDNGL_WREPROMERROR)) {
					
                    /* Ignore errors here */
                    smrtdnglWaitForTwcFlag(
										   SMARTDNGL_TWC_TIMEOUT, dongle, control);
					
                    continue;
                } else {
                    break;
                }
            }
			
            /* Wait here for Twc (Time Write cycle) to reset */
            smrtdnglWaitForTwcFlag(SMARTDNGL_TWC_TIMEOUT, dongle, control);
            if (control->status != SMARTDNGL_ERR_OK) {
                break;
            }
			
            break;
        }
		
        /*
		 Error from retry loop: break here and fail
		 */
        if (control->status != SMARTDNGL_ERR_OK) {
            break;
        }
    }
	
FAIL:
    if (memory != NULL) {
        free(memory);
    }
	
    savedStatus = control->error;
	
    /* Attempt to Lock SmartDongle. */
    smrtdnglSendSecureStateReset(dongle, control);
	
    control->error = savedStatus;
    return;
}

unsigned long long SmartDongleGetKeylessChallenge() {
	unsigned long long challenge = 0;
	smartdngl_control control;
    smartdngl_dev* dongle;
	
    dongle = smrtdnglGetDev(&control);
    if (dongle == NULL) {
		
        control.status = 0;
        smrtdnglProbe(&control);
		
        dongle = smrtdnglGetDev(&control);
        if (dongle == NULL)
        {
            return control.error;
        }
    }
	unsigned char readErrorMask = SMARTDNGL_RDEPROMERROR;
    int savedError;
	
	
    /*
	 Reset key state machine here for all but v1.50 or higher SmartDongles
	 */
    if (!(dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT)) {
		
        smrtdnglResetSdState(dongle, &control);
        if (control.error != SMARTDNGL_ERR_OK) {
            return -1;
        }
		
        smrtdnglWaitForSdState(
							   SMARTDNGL_STATE0_TIMEOUT,
							   SMARTDNGL_STATE0,
							   dongle,
							   &control);
        if (control.error != SMARTDNGL_ERR_OK) {
            return -1;
        }
    }
	
    /*
	 Request SmartDongle create a challenge.
	 */
    smrtdnglSendChallengeRequest(dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        return -1;
    }
	
    /*
	 SmartDongle challenge is ready when state is SMARTDNGL_STATE1 (1)
	 */
    smrtdnglWaitForSdState(
						   SMARTDNGL_STATE1_TIMEOUT,
						   SMARTDNGL_STATE1,
						   dongle,
						   &control);
    if (control.error != SMARTDNGL_ERR_OK) {
		
        /*
		 Was there a fatal EPROM read error?
		 Ignore readPageFlags IOCTL error
		 */
        savedError = control.error;
		
        smrtdnglReadPageFlags(dongle, &control);
        if (control.error == SMARTDNGL_ERR_OK) {
			
            if (control.status & readErrorMask) {
                control.error = SMARTDNGL_ERR_RD_EPROM;
            } else {
                control.error = savedError;
            }
        } else {
            control.error = savedError;
        }
		
        return -1;
    }
	
    /*
	 Get the SmartDongle's challenge.
	 */
    smrtdnglGetKey(dongle, &control, &challenge);
    if (control.error != SMARTDNGL_ERR_OK) {
        return -1;
    }
	
	return challenge;
}

int SmartDongleSendKeylessResponse(unsigned long long response) {
	unsigned long long challengeResponse = response;
	smartdngl_control control;
    smartdngl_dev* dongle;
	
    dongle = smrtdnglGetDev(&control);
    if (dongle == NULL) {
		
        control.status = 0;
        smrtdnglProbe(&control);
		
        dongle = smrtdnglGetDev(&control);
        if (dongle == NULL)
        {
            return control.error;
        }
    }
	
	smrtdnglSendReply(dongle, &control, &challengeResponse);
    if (control.error != SMARTDNGL_ERR_OK) {
        return control.error;
    }
	
    /*
	 When the state sequences back to 0
	 the host response has been verified.
	 */
    smrtdnglWaitForSdState(
						   SMARTDNGL_STATE0_TIMEOUT,
						   SMARTDNGL_STATE0,
						   dongle,
						   &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        return control.error;
    } /*
	   Is the host verified, dongle ENABLED?
	   */
    smrtdnglReadSecurityState(dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        return control.error;
    }
	
    if (control.status != SMARTDNGL_UNLOCKED) {
        control.error = SMARTDNGL_ERR_VALIDATE;
        return control.error;
    }
	
	return control.error;
}

/*
    Unlock SmartDongle with primes P1 and P2.
*/
static void smrtdnglEnable(
    unsigned long long *P1,
    unsigned long long *P2,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char readErrorMask = SMARTDNGL_RDEPROMERROR;
    unsigned long long X;
    int savedError;
    int i;
    int numLoops;
    char sn[SMARTDNGL_SN_LENGTH * 10 + 1];

    getSerialNumber(sn, dongle, control);
    numLoops = getLoopCount(sn);

    /*
        Reset key state machine here for all but v1.50 or higher SmartDongles
    */
    if (!(dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT)) {

        smrtdnglResetSdState(dongle, control);
        if (control->error != SMARTDNGL_ERR_OK) {
            return;
        }

        smrtdnglWaitForSdState(
            SMARTDNGL_STATE0_TIMEOUT,
            SMARTDNGL_STATE0,
            dongle,
            control);
        if (control->error != SMARTDNGL_ERR_OK) {
            return;
        }
    }

    /*
        Request SmartDongle create a challenge.
    */
    smrtdnglSendChallengeRequest(dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    /*
        SmartDongle challenge is ready when state is SMARTDNGL_STATE1 (1)
    */
    smrtdnglWaitForSdState(
        SMARTDNGL_STATE1_TIMEOUT,
        SMARTDNGL_STATE1,
        dongle,
        control);
    if (control->error != SMARTDNGL_ERR_OK) {

        /*
            Was there a fatal EPROM read error?
            Ignore readPageFlags IOCTL error
        */
        savedError = control->error;

        smrtdnglReadPageFlags(dongle, control);
        if (control->error == SMARTDNGL_ERR_OK) {

            if (control->status & readErrorMask) {
                control->error = SMARTDNGL_ERR_RD_EPROM;
            } else {
                control->error = savedError;
            }
        } else {
            control->error = savedError;
        }

        return;
    }

    /*
        Get the SmartDongle's challenge.
    */
    smrtdnglGetKey(dongle, control, &X);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    for(i=0; i<numLoops; i++){
        X *= *P1;
        X += *P2;
    }

    /*
        Reply to challenge
    */
    smrtdnglSendReply(dongle, control, &X);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    /*
        When the state sequences back to 0
        the host response has been verified.
    */
    smrtdnglWaitForSdState(
        SMARTDNGL_STATE0_TIMEOUT,
        SMARTDNGL_STATE0,
        dongle,
        control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    } /*
        Is the host verified, dongle ENABLED?
    */
    smrtdnglReadSecurityState(dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    if (control->status != SMARTDNGL_UNLOCKED) {
        control->error = SMARTDNGL_ERR_VALIDATE;
        return;
    }

    /*
        Force a v1.50 or greater SmartDongle to identify itself.
    */
    if (dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT) {
        smrtdnglValidateUsk(P1, P2, dongle, control);
    }

    return;
}

/*
    For SmartDongle v1.50, force a SmartDongle to reply to a host challenge.
    Useful if a user does not write any secrets to SmartDongle's user memory
    and simply checks for presence of a SmartDongle.  An attacker could spoof
    a protected application program by using a modified driver or usk.dll

    SmartDongle must be enabled to support a host challenge.
*/
static void smrtdnglValidateUsk(
    unsigned long long *P1,
    unsigned long long *P2,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned long long X;
    unsigned long long expected;
    char sn[SMARTDNGL_SN_LENGTH * 10 + 1];
    int numLoops;
    int i;


    srandom(time(NULL));

    X = (unsigned long long)random();
    X |= (unsigned long long)random() << 16;
    X |= (unsigned long long)random() << 32;
    X |= (unsigned long long)random() << 48;

    /* High bit must be set */
    X |= 0x8000000000000000ULL;

    getSerialNumber(sn, dongle, control);
    numLoops = getLoopCount(sn);

    /* Calculate expected reply from SmartDongle */
    expected = X;
    for(i=0; i<numLoops; i++){
        expected *= *P1;
        expected += *P2;
    }

    /*
        Send challenge.
        Assuming key state was set to 0 at termination of
        SmartDongle enable sequence.
    */
    smrtdnglSendChallenge(dongle, control, &X);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    /*
        Wait for SmartDongle's reply to host challenge.
    */
    smrtdnglWaitForSdState(
        SMARTDNGL_STATE1_TIMEOUT,
        SMARTDNGL_STATE1,
        dongle,
        control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    /*
        Get the SmartDongle's reply to challenge.
    */
    smrtdnglGetKey(dongle, control, &X);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    if (expected != X) {
        control->error = SMARTDNGL_ERR_USK_VALIDATE;
    }

    return;
}

/*
    Send reply calculated with a one-way function from the challenge.
    SmartDongle will unlock if reply is correct.
*/
static void smrtdnglSendReply(
    smartdngl_dev *dongle,
    smartdngl_control *control,
    unsigned long long *X)
{
    int sent;


    control->error = SMARTDNGL_ERR_OK;

    /*
        If signed read/write is supported
    */
    if (dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT) {

        dongle->sig = (unsigned short)((*X & 0xFFFF000000000000ULL)>>48);
        dongle->sig2 = (unsigned short)((*X & 0x0000FFFF00000000ULL)>>32);

        *X &= 0xFFFFFFFFULL;
    }

    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_REPLY,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        X,
        SMARTDNGL_KEYSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < SMARTDNGL_KEYSIZE) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_REPLY;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    Send challenge to v1.50 SmartDongle.
*/
static void smrtdnglSendChallenge(
    smartdngl_dev *dongle,
    smartdngl_control *control,
    unsigned long long *X)
{
    int sent;


    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_REPLY,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        X,
        SMARTDNGL_KEYSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < SMARTDNGL_KEYSIZE) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_REPLY;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    Bi-color LED  Control.
*/
static void smrtdnglLedControl(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;
    unsigned char data[SMARTDNGL_PACKETSIZE];

    control->error = SMARTDNGL_ERR_OK;
    data[dongle->byte0] = SMARTDNGL_LEDPORT;
    data[dongle->byte1] = control->status; /* color */

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_LED,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < SMARTDNGL_PACKETSIZE) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_LED;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    Request SmartDongle to issue a challenge.
    State will sequence to SMARTDNGL_STATE0 when challenge is ready.
    Smartdongle will unlock with a correct reply to the challenge.
*/
static void smrtdnglSendChallengeRequest(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;


    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_CHALLENGE_REQ,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < 0) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_REQ_CHLG;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    Resets SmartDongle Request/Challenge state.
*/
static void smrtdnglResetSdState(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;


    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_STATE_RESET,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < 0) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RESET_STATE;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    Locks SmartDongle
*/
static void smrtdnglSendSecureStateReset(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;


    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_SECURITY_RESET,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < 0) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_LOCK;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    Send a page of data out to the SmartDongle's write buffer, call
    smrtdnglReadPageFlags to poll the status of the EEPROM write buffer.
*/
static void smrtdnglSendEepromPage(
    unsigned short address,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;
    int size;
    char *iobuffer;
    unsigned short sig, sig2;
    unsigned char data[4 * sizeof(unsigned short) + SMARTDNGL_EPROMPGSZ];


    iobuffer = dongle->iobuffer;
    control->error = SMARTDNGL_ERR_OK;

    /*
        32 bit Signature + EEPROM address + 16 bytes of data
    */
    if (dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT) {
        sig  =  sigFlip(dongle->sig, dongle);
        sig2 =  sigFlip(dongle->sig2, dongle);
        address = sigFlip(address, dongle);

        memcpy(data, &sig, sizeof(unsigned short));
        memcpy(data+sizeof(unsigned short), &sig2, sizeof(unsigned short));
        memcpy(data+(2*sizeof(unsigned short)),&address, sizeof(unsigned short));
        memcpy(data+(3*sizeof(unsigned short)), iobuffer, SMARTDNGL_EPROMPGSZ);

        size = 4*sizeof(unsigned short) + SMARTDNGL_EPROMPGSZ;
    }

    /*
        EEPROM address + 16 bytes of data
    */
    else {
        memcpy(data, &address, sizeof(unsigned short));
        memcpy(data+sizeof(unsigned short), iobuffer, SMARTDNGL_EPROMPGSZ);

        size = sizeof(unsigned short) + SMARTDNGL_EPROMPGSZ;
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_EPROM_PAGE,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        size,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < size) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_WR_EPROM;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    Request SmartDongle to prefetch EPROM pages from given address.
*/
static void smrtdnglRequestEepromPrefetch(
    unsigned short address,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;


    control->error = SMARTDNGL_ERR_OK;

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_EPROM_PAGE_REQ,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        address,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < 0) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_REQ_PAGE;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}


#if !defined(SMARTDONGLE_JNI)
/*
    Sofware reset for SmartDongle
*/
static void smrtdnglSendSoftReset(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;


    control->error = SMARTDNGL_ERR_OK;

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_SOFT_RESET,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    /* SmartDongle is reset, so error will always occurr, ignore */
    return;
}

/*
    Overwrite the user's EPROM encryption mode flags.  Note v1.50 SmartDongle
    must be reset prior to overwriting mode bits, this will breaks the libusb
    interface so this command should not be used with v1.50 SmartDongles
*/
static void smrtdnglSendEpromMode(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;


    control->error = SMARTDNGL_ERR_OK;

    sent =  smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_SET_MODE,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        control->status,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < SMARTDNGL_PACKETSIZE) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_WR_MODE;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}

/*
    For SmartDongle >=v1.50 using >= v2.3 driver only.
    This will configure the SmartDongle for Strict Access.
    A new key and 32 bit signatures are required for each user
    read/write transaction against user memory.

    Caution: This configuration is permanent.
                SmartDongle must be sent back to the manufacturer to
                be reprogrammed to use the older drivers (<= v2.2.0).
*/
static void smrtdnglSendSetStrictOnly(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int sent;


    control->error = SMARTDNGL_ERR_OK;

    /*
        Requires >=1.50 SmartDongles
    */
    if (!(dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT)) {
        control->error = SMARTDNGL_ERR_REQUIRES_1_50;
        return;
    }

    /*
        Keep customer from burning memory with repeated requests.
        Return if SmartDongle previously set to strict access.
    */
    if (dongle->opt & SMARTDNGL_OPTION_STRICT_ONLY) {
        return;
    }

    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_SET_STRICT_ONLY,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (sent < 0) {
        if (sent == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_SET_STRICT;
        }
        else {
            control->error = -sent;
        }
    }

    return;
}
#endif


/*
    Retrieve either the challenge from SmartDongle or (v1.50)  Smartdongle's
    reply to the host challenge.
*/
static void smrtdnglGetKey(
    smartdngl_dev  *dongle,
    smartdngl_control *control,
    unsigned long long *X)
{
    int recieved;


    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    recieved = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_CHALLENGE,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        X,
        SMARTDNGL_KEYSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_KEYSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_CHALLENGE;
        }
        else {
            control->error = -recieved;
        }
    }

    return;
}

/*
    Read the locked/unlocked state.
*/
static void smrtdnglReadSecurityState(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char data[SMARTDNGL_PACKETSIZE];
    int recieved;


    control->error = SMARTDNGL_ERR_OK;

    recieved = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_SECURITY_STATE,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_PACKETSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_STATE;
        }
        else {
            control->error = -recieved;
        }
    }

    /* state in first byte */
    control->status = data[dongle->byte0];

    return;
}

/*
    Read the state of authentication state machine.
*/
static void smrtdnglReadSdState(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char data[SMARTDNGL_PACKETSIZE];
    int recieved;

    control->error = SMARTDNGL_ERR_OK;

    recieved = smrtdnglSendControlMessage(
        dongle,
       SMARTDNGL_GET_SD_STATE,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_PACKETSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_STATE;
        }
        else {
            control->error = -recieved;
        }
    }

    /* State in first byte */
    control->status = data[dongle->byte0];

    return;
}

/*
    If successful, a postive number returned of the number of bytes read.
*/
static void smrtdnglGetEpromPage(
    int *size,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int recieved;
    unsigned short sig, sig2;
    char *iobuffer;


    iobuffer = dongle->iobuffer;

    /*
        If the SmartDongle supports using a "signed" USB request block.
    */
    if (dongle->opt & SMARTDNGL_OPTION_EXTENDED_SUPPORT) {
        sig = dongle->sig;
        sig2 = dongle->sig2;
    }
    else {
        sig = 0;
        sig2 = 0;
    }

    recieved = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_EPROM_PAGE,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        sig2,
        sig,
        iobuffer,
        SMARTDNGL_MAXPREFETCH,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < 0) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_EPROM;
        }
        else {
            control->error = -recieved;
        }

        *size = 0;
        return;
    }

    *size = recieved;
    return;
}

/*
    Fetch bitmapped byte that indicates when a EPROM page is ready
    or if a read/write error occurred.
*/
static void smrtdnglReadPageFlags(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char data[SMARTDNGL_PACKETSIZE];
    int recieved;


    control->error = SMARTDNGL_ERR_OK;

    recieved = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_EPROM_PAGE_ERROR,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_PACKETSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_PGFLGS;
        }
        else {
            control->error = -recieved;
        }
    }

    /* EPROM access flags in first byte */
    control->status = data[dongle->byte0];

    return;
}

/*
    Read the Time Write Cycle status.
*/
static void smrtdnglReadTwcStatus(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char data[SMARTDNGL_PACKETSIZE];
    int recieved;


    control->error = SMARTDNGL_ERR_OK;

    recieved  = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_TWC_STATE,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_PACKETSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_TWC;
        }
        else {
            control->error = -recieved;
        }
    }

    /* Status in first byte */
    control->status = data[dongle->byte0];

    return;
}

/*
    Read the status of SmartDongle task loop.
*/
static void smrtdnglReadBusyStatus(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char data[SMARTDNGL_PACKETSIZE];
    int recieved;


    control->error = SMARTDNGL_ERR_OK;

    recieved  = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_BUSY_STATE,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_PACKETSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_BSY;
        }
        else {
            control->error = -recieved;
        }
    }

    /* Busy status in first byte. */
    control->status = data[dongle->byte0];

    return;
}

/*
    SmartDongle option bits, for extended functionality.
*/
static void smrtdnglGetOpt(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char data[SMARTDNGL_PACKETSIZE];
    int recieved;


    /* clear options */
    dongle->opt = 0;

    /*
        Wait for task_loop processing to finish
         control->error is reset in smrtdnglReadBusyStatus().
    */
    smrtdnglWaitForBusyStatus(SMARTDNGL_BSY_TIMEOUT, dongle, control);
    if (control->error != SMARTDNGL_ERR_OK) {
        return;
    }

    recieved = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_OPTIONS,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_PACKETSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_OPTIONS;
        }
        else {
            control->error = -recieved;
        }
    }

    /*
        Map firmware flags to option bitmaps
        SmartDongle options in first byte
    */
    if (data[dongle->byte0]  & SMARTDNGL_FIRMWARE_EXTENDED_SUPPORT2) {
        dongle->opt |= SMARTDNGL_OPTION_EXTENDED_SUPPORT2;
    }

    if ( data[dongle->byte0]  & SMARTDNGL_FIRMWARE_EXTENDED_SUPPORT) {
        dongle->opt |= SMARTDNGL_OPTION_EXTENDED_SUPPORT;
    }

    if ( data[dongle->byte0]  & SMARTDNGL_FIRMWARE_STRICT_ONLY) {
        dongle->opt |= SMARTDNGL_OPTION_STRICT_ONLY;
    }

    return;
}

/*
    Read the user's EPROM encryption mode flags.
*/
static void smrtdnglReadCryptMode(
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    unsigned char data[SMARTDNGL_PACKETSIZE];
    int recieved;


    control->error = SMARTDNGL_ERR_OK;

    recieved = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_GET_MODE,
        USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        data,
        SMARTDNGL_PACKETSIZE,
        SMARTDNGL_MESG_TIMEOUT);

    /*
        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (recieved < SMARTDNGL_PACKETSIZE) {
        if (recieved == -SMARTDNGL_ERR_USB_CONTROL_MSG) {
            control->error = SMARTDNGL_ERR_RD_MODE;
        }
        else {
            control->error = -recieved;
        }
    }

    /*
        User space encryption mode in first byte
        Right justify the value. Mask off extraneous bits.
    */
    control->status = data[dongle->byte0];
    control->status >>= 4;
    control->status &= SMARTDNGL_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE|
        SMARTDNGL_OPTION_ON_CHIP_EPROM_PROTECT_FAST;

    return;
}

/*
    Poll EPROM flags that indicate a page is ready or a read error occurred.
*/
static void smrtdnglWaitForPageRead(
    int timeout,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int time = timeout / SMARTDNGL_PAGE_TIMEOUT_POLL;
    int i;


    /* Ready if EPROM page valid bit set or any EPROM read error. */
    unsigned char ready =
        SMARTDNGL_RDPAGEVALID|
        SMARTDNGL_RDADDRESS|
        SMARTDNGL_RDDENY|
        SMARTDNGL_RDEPROMERROR;

    for (i=0; i< SMARTDNGL_PAGE_TIMEOUT_POLL; i++) {

        smrtdnglReadPageFlags(dongle, control);
        if (control->error == SMARTDNGL_ERR_OK) {
            /*
                Break if EEPROM page ready or error has occurred.
            */
            control->status &= ready;
            if (control->status) {
                break;
            }
        }

        usleep(time);
    }

    /*
        Timeout error.

        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (i == SMARTDNGL_POLL_RETRYS) {
         if (control->error == SMARTDNGL_ERR_OK) {
            control->error = SMARTDNGL_ERR_EPROM_RD_TIMEOUT;
        }
    }

    return;
}

/*
    Poll busy status to see if SmartDongle running function in task loop.
*/
static void smrtdnglWaitForBusyStatus(
    int timeout,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int time = timeout / SMARTDNGL_POLL_RETRYS;
    int i;


    for (i=0; i<SMARTDNGL_POLL_RETRYS; i++) {

        smrtdnglReadBusyStatus(dongle, control);

        if (control->error == SMARTDNGL_ERR_OK) {
            /* Break if not busy */
            if (control->status == SMARTDNGL_BUSY0) {
                break;
            }
        }

        usleep(time);
    }

    /*
        Timeout error

        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (i == SMARTDNGL_POLL_RETRYS) {
         if (control->error == SMARTDNGL_ERR_OK) {
            control->error = SMARTDNGL_ERR_BSY_TIMEOUT;
        }
    }

    return;
}

/*
    Poll dongle state to see if challenge is ready or reply has been processed
*/
static void smrtdnglWaitForSdState(
    int timeout,
    unsigned char state,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int time = timeout / SMARTDNGL_POLL_RETRYS;
    int i;


    for (i = 0; i < SMARTDNGL_POLL_RETRYS; i++) {

        smrtdnglReadSdState(dongle, control);

        if (control->error == SMARTDNGL_ERR_OK) {
            /* Break if desired state */
            if (control->status == state) {
                break;
            }
        }

        usleep (time);
    }

    /*
        Timeout error.

        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (i == SMARTDNGL_POLL_RETRYS) {
         if (control->error == SMARTDNGL_ERR_OK) {
            control->error = SMARTDNGL_ERR_SM_TIMEOUT;
        }
    }

    return;
}

/*
    Serial Peripheral Interface (SPI) EPROM's require 5mS Time Write Cycle
    (TWC) after each page write.
    SmartDongle has a local timer driven by the 1mS interrupt and will deny
    EPROM access while the timer is running.

    The TWC state will toggle to 0 (SMARTDNGL_TWCINACTIVE) on inactive timer.
*/
static void smrtdnglWaitForTwcFlag(
    int timeout,
    smartdngl_dev *dongle,
    smartdngl_control *control)
{
    int time = timeout / SMARTDNGL_POLL_RETRYS;
    int i;


    for (i = 0; i < SMARTDNGL_POLL_RETRYS; i++) {

        smrtdnglReadTwcStatus(dongle, control);

        if (control->error == SMARTDNGL_ERR_OK) {
            /* Break on TWC inactive */
            if (control->status == SMARTDNGL_TWC0) {
                break;
            }
        }

        usleep(time);
    }

    /*
        Timeout error.

        Allow libusb errors from commands usb_open() and
        usb_claim_interface() to be returned to user.
    */
    if (i == SMARTDNGL_POLL_RETRYS) {
         if (control->error == SMARTDNGL_ERR_OK) {
            control->error = SMARTDNGL_ERR_TWC_TIMEOUT;
        }
    }

    return;
}


/*
    Return error string corresponding to error code.
    User must allocate string size of SMARTDNGL_ERROR_STRING_LENGTH_MAX
*/
void SmartDongleGetErrorString(char *errorString, int error)
{
    switch (error)
    {

        case SMARTDNGL_ERR_OK:
            strcpy(errorString, "no errors");
            break;

        case SMARTDNGL_ERR_MEM:
            strcpy(errorString, "System memory allocation error");
            break;

        case SMARTDNGL_ERR_EPROM_SIZE:
            strcpy(
                errorString,
                "Address offset or buffer size exceeds EPROM memory");
            break;


        /*
            Timeout for State Machine
        */
        case SMARTDNGL_ERR_SM_TIMEOUT:
            strcpy(errorString, "Timeout waiting for SmartDongle State change");
            break;

        /*
            Timeout waiting for TWC to go to "inactive" state
        */
        case SMARTDNGL_ERR_TWC_TIMEOUT:
            strcpy(errorString, "Timeout waiting for Time Write Cycle");
            break;

        /*
            Timeout waiting for Task Loop to finish
        */
        case SMARTDNGL_ERR_BSY_TIMEOUT:
            strcpy(errorString, "Timeout waiting for Busy SmartDongle");
            break;

        /*
            Timeout Waiting For Read Flag(s) to indicate either Page Ready or
            error.
        */
        case SMARTDNGL_ERR_EPROM_RD_TIMEOUT:
            strcpy(errorString, "Timeout waiting for Read Flag(s)");
            break;

        case SMARTDNGL_ERR_VALIDATE:
            strcpy(
                errorString,
                "Host failed to validate against SmartDongle keys");
            break;

        case SMARTDNGL_ERR_USK_VALIDATE:
            strcpy(
                errorString,
                "SmartDongle failed to validate against user keys");
            break;

        case SMARTDNGL_ERR_BUFSIZE:
            strcpy(errorString, "Insufficient buffer size for command");
            break;

        case SMARTDNGL_ERR_DATA_FORMAT:
            strcpy(errorString, "Data format Error");
            break;

        case SMARTDNGL_ERR_BAD_ARG:
            strcpy(errorString, "One or more arguments are not correct");
            break;

        case SMARTDNGL_ERR_NODEV:
            strcpy(errorString, "No SmartDongles found");
            break;

        /*
            For USB Vendor Device Request control failure.
        */
        case SMARTDNGL_ERR_CHALLENGE:
        case SMARTDNGL_ERR_REPLY:
        case SMARTDNGL_ERR_REQ_CHLG:
        case SMARTDNGL_ERR_REQ_PAGE:
        case SMARTDNGL_ERR_RD_STATE:
        case SMARTDNGL_ERR_RD_STATUS:
        case SMARTDNGL_ERR_RD_BSY:
        case SMARTDNGL_ERR_RD_TWC:
        case SMARTDNGL_ERR_RD_PGFLGS:
        case SMARTDNGL_ERR_RD_EPROM:
        case SMARTDNGL_ERR_WR_EPROM:
        case SMARTDNGL_ERR_RESET_SM:
        case SMARTDNGL_ERR_RESET_STATE:
        case SMARTDNGL_ERR_LOCK:
        case SMARTDNGL_ERR_SOFT_RST:
        case SMARTDNGL_ERR_LED:
        case SMARTDNGL_ERR_RD_MODE:
        case SMARTDNGL_ERR_WR_MODE:
        case SMARTDNGL_ERR_SET_STRICT:
        case SMARTDNGL_ERR_RD_OPTIONS:
            sprintf(
                errorString,
                "USB Vendor Device Request control failure: %d",
                error);
            break;

        case SMARTDNGL_ERR_OPEN:
            strcpy(errorString, "Failed to open SmartDongle");
            break;

        case SMARTDNGL_ERR_GET_SN:
            strcpy(errorString, "Error reading Serial Number");
            break;

        case SMARTDNGL_ERR_DEV_MAX:
            strcpy(errorString, "Too many SmartDongles plugged in");
            break;

        /*
            Error attempting to use extended device functionality with old
            SmartDongle.
        */
        case SMARTDNGL_ERR_REQUIRES_1_50:
            strcpy(
                errorString,
                "This command requires a version 1.50 Smartdongle");
            break;

        case SMARTDNGL_ERR_REQUIRES_1_70:
            strcpy(
                errorString,
                "This command requires a version 1.70 Smartdongle");
            break;

            case SMARTDNGL_ERR_USB_OPEN:
            strcpy(errorString, "libusb command usb_open() failed");
            break;


        /*
            The libusb command usb_claim_interface() failed

            Error result from:

            1) The user has insufficient privileges to read files under
               /proc/bus/usb

            2) SmartDongle interface  may have been claimed by another process
               or thread.

            3) SmartDongle been unplugged.
        */
        case SMARTDNGL_ERR_USB_CLAIM_INTERFACE:
            strcpy(errorString, "libusb command usb_claim_interface() failed");
            break;

        case SMARTDNGL_ERR_USB_CONTROL_MSG:
            strcpy(errorString, "The libusb_control_msg() failed");
            break;

        /*
            Unknown.
        */
        case SMARTDNGL_ERR_UNKNOWN:
        default:
            strcpy (errorString, "An unknown error occurred");
            break;
    }
}


/*
    flips a short to dongle endianness
*/
static unsigned short sigFlip(unsigned short sig, smartdngl_dev *dongle) {
   if (dongle->big_e) {
      sig = ((sig >> 8) & 0xFF) | ((sig << 8) & 0xFF00);
   }

   return sig;
}

/*
    converts a single integer from one endianness to another
*/
static void intFlip(
   unsigned char *c)
{
   char temp;

   temp = c[0];
   c[0] = c[7];
   c[7] = temp;
   temp = c[1];
   c[1] = c[6];
   c[6] = temp;
   temp = c[2];
   c[2] = c[5];
   c[5] = temp;
   temp = c[3];
   c[3] = c[4];
   c[4] = temp;

}

/*
    switches endianness of the data passed to it.
*/
static void endianConvert(
   unsigned char *data,
   unsigned int size)
{
   int i;

   if(size < SMARTDNGL_PACKETSIZE) {
      return;
   }

   for(i = 0; i < size; i += SMARTDNGL_PACKETSIZE) {
      intFlip(&data[i]);
   }
   return;
}



#if defined(SMARTDONGLE_JNI)

/*
   Given the correct keys, p1 and p2, will place the dongle into
   the enabled state and read EEPROM into buffer.
*/
JNIEXPORT jint JNICALL SED_XXXRead(
    JNIEnv *env,
    jclass cls,
    jlong p1,
    jlong p2,
    jint address,
    jbyteArray arr,
    jint size)
{
    smartdngl_control control;
    smartdngl_dev* dongle;
    int status;
    jbyte *carr;

    carr = (*env)->GetByteArrayElements(env, arr, NULL);
    if (carr == NULL) {
        /* exception occurred */
        lastError = SMARTDNGL_ERR_UNKNOWN;
        return SMARTDNGL_ERR_UNKNOWN;
    }

    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {

        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL) {
            (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
            lastError = control.error;
            return control.error;
        }
    }

    /*
        Attempt to read user EEPROM space.
        On failure, attempt to disable dongle and return error.
    */
    smrtdnglRead(
        (unsigned long long*)&p1,
        (unsigned long long*)&p2,
        (unsigned short)address,
        (unsigned char*)carr,
        size,
        dongle,
        &control);
    if (control.error != SMARTDNGL_ERR_OK) {

        status = control.error;
        smrtdnglSendSecureStateReset(dongle, &control);
        (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
        lastError = status;
        return status;
    }

    smrtdnglSendSecureStateReset(dongle, &control);
    (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
    return control.error;
}

JNIEXPORT jint JNICALL SED_XXXKeylessRead(
								   JNIEnv *env,
								   jclass cls,
								   jint address,
								   jbyteArray arr,
								   jint size)
{
    smartdngl_control control;
    smartdngl_dev* dongle;
    int status;
    jbyte *carr;
	
    carr = (*env)->GetByteArrayElements(env, arr, NULL);
    if (carr == NULL) {
        /* exception occurred */
        lastError = SMARTDNGL_ERR_UNKNOWN;
        return SMARTDNGL_ERR_UNKNOWN;
    }
	
    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {
		
        smrtdnglProbe(&control);
		
        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL) {
            (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
            lastError = control.error;
            return control.error;
        }
    }
	
    /*
	 Attempt to read user EEPROM space.
	 On failure, attempt to disable dongle and return error.
	 */
    smrtdnglKeylessRead(
				 (unsigned short)address,
				 (unsigned char*)carr,
				 size,
				 dongle,
				 &control);
    if (control.error != SMARTDNGL_ERR_OK) {
		
        status = control.error;
        smrtdnglSendSecureStateReset(dongle, &control);
        (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
        lastError = status;
        return status;
    }
	
    smrtdnglSendSecureStateReset(dongle, &control);
    (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
    return control.error;
}


/*
    Given the correct keys, p1 and P2, will place the dongle into
    the enabled state and write buffer into EEPROM.
*/
JNIEXPORT jint JNICALL SED_XXXWrite(
    JNIEnv *env,
    jclass cls,
    jlong p1,
    jlong p2,
    jint address,
    jbyteArray arr,
    jint size)
{
    smartdngl_control control;
    smartdngl_dev *dongle;
    int status;
    jbyte *carr;


    carr = (*env)->GetByteArrayElements(env, arr, NULL);
    if (carr == NULL) {
        /* exception occurred */
        lastError = SMARTDNGL_ERR_UNKNOWN;
        return SMARTDNGL_ERR_UNKNOWN;
    }

    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {

        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL) {
            (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
            lastError = control.error;
            return control.error;
        }
    }

    /*
        Attempt to write to user EEPROM space.
        On failure, attempt to disable dongle and return error.
    */
    smrtdnglWrite(
        (unsigned long long*)&p1,
        (unsigned long long*)&p2,
        (unsigned short)address,
        (unsigned char*)carr,
        size,
        dongle,
        &control);
    if (control.error != SMARTDNGL_ERR_OK)
    {

        status = control.error;
        smrtdnglSendSecureStateReset(dongle, &control);
        (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
        lastError = status;
        return status;
    }

    smrtdnglSendSecureStateReset(dongle, &control);
    (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
    return control.error;
}

JNIEXPORT jint JNICALL SED_XXXKeylessWrite(
									JNIEnv *env,
									jclass cls,
									jint address,
									jbyteArray arr,
									jint size)
{
    smartdngl_control control;
    smartdngl_dev *dongle;
    int status;
    jbyte *carr;
	
	
    carr = (*env)->GetByteArrayElements(env, arr, NULL);
    if (carr == NULL) {
        /* exception occurred */
        lastError = SMARTDNGL_ERR_UNKNOWN;
        return SMARTDNGL_ERR_UNKNOWN;
    }
	
    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {
		
        smrtdnglProbe(&control);
		
        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL) {
            (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
            lastError = control.error;
            return control.error;
        }
    }
	
    /*
	 Attempt to write to user EEPROM space.
	 On failure, attempt to disable dongle and return error.
	 */
    smrtdnglKeylessWrite(
				  (unsigned short)address,
				  (unsigned char*)carr,
				  size,
				  dongle,
				  &control);
    if (control.error != SMARTDNGL_ERR_OK)
    {
		
        status = control.error;
        smrtdnglSendSecureStateReset(dongle, &control);
        (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
        lastError = status;
        return status;
    }
	
    smrtdnglSendSecureStateReset(dongle, &control);
    (*env)->ReleaseByteArrayElements(env, arr, carr, 0);
    return control.error;
}

JNIEXPORT jlong JNICALL SED_XXXGetKeylessChallenge(
								JNIEnv *env, jclass cls) {
    unsigned long long ret = SmartDongleGetKeylessChallenge();
	return (jlong)ret;
}

JNIEXPORT jint JNICALL SED_XXXSendKeylessResponse(
								JNIEnv *env, jclass cls, jlong response) {
	int ret = SmartDongleSendKeylessResponse(response);
	return (jint)ret;
}

/*
    Switch SmartDongle LED indicator to Green.
*/
JNIEXPORT jint JNICALL SED_XXXLedGreen(
    JNIEnv *env , jclass cls)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {

        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL)
        {
            lastError = control.error;
            return control.error;
        }
    }

    control.status = SMARTDNGL_GREEN;
    smrtdnglLedControl(dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        lastError = control.error;
    }

    return control.error;
}


/*
    Switch SmartDongle LED indicator to Red.
*/
JNIEXPORT jint JNICALL SED_XXXLedRed(
    JNIEnv *env , jclass cls)
{
    smartdngl_control control;
    smartdngl_dev* dongle;
    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {
        smrtdnglProbe(&control);
        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL) {
            lastError = control.error;
            return control.error;
        }
    }
    control.status = SMARTDNGL_RED;
    smrtdnglLedControl(dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        lastError = control.error;
    }
    return control.error;
}


/*
    Switch SmartDongle LED indicator to Off.
*/
JNIEXPORT jint JNICALL SED_XXXLedOff(
    JNIEnv *env , jclass cls)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {

        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL)
        {
            lastError = control.error;
            return control.error;
        }
    }

    control.status = SMARTDNGL_OFF;
    smrtdnglLedControl(dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        lastError = control.error;
    }

    return control.error;
}


/*
    SmartDongle software reset
*/
JNIEXPORT jint JNICALL SED_XXXReset(
    JNIEnv *env, jclass cls)
{
    smartdngl_control control;
    smartdngl_dev* dongle;
    int sent;


    dongle = smrtdnglGetDev (&control);
    if (dongle==NULL) {

        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev (&control);
        if (dongle==NULL)
        {
            lastError = control.error;
            return control.error;
        }
    }

    sent = smrtdnglSendControlMessage(
        dongle,
        SMARTDNGL_SEND_SOFT_RESET,
        USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
        0,
        0,
        NULL,
        0,
        SMARTDNGL_MESG_TIMEOUT);

    if (sent < 0) {
        lastError = SMARTDNGL_ERR_SOFT_RST;
        control.error = SMARTDNGL_ERR_SOFT_RST;
    }

    return control.error;
}


/*
    For SmartDongle version >=1.42

    Set the current user memory space encryption mode.
*/
JNIEXPORT jbyte JNICALL SED_XXXReadCryptMode(
    JNIEnv *env, jclass cls)
{
    smartdngl_control control;
    smartdngl_dev* dongle;


    dongle = smrtdnglGetDev(&control);
    if (dongle==NULL) {

        smrtdnglProbe(&control);

        dongle = smrtdnglGetDev(&control);
        if (dongle==NULL)
        {
            lastError = control.error;
            return control.error;
        }
    }

    smrtdnglReadCryptMode(dongle, &control);
    if (control.error != SMARTDNGL_ERR_OK) {
        lastError = control.error;
    }

    return control.status;
}

JNIEXPORT void JNICALL SED_XXXFileDecrypt(JNIEnv *env, jclass cls,
	jbyteArray secret, jint scrtLen, jbyteArray key, jint blockLen) {
	jbyte *ndata;
	ndata = (*env)->GetByteArrayElements(env, secret, NULL);
	jbyte *myKey;
	myKey = (*env)->GetByteArrayElements(env, key, NULL);
	JIKpbm3((unsigned char*)ndata, scrtLen, (unsigned char*)myKey, blockLen);
	(*env)->ReleaseByteArrayElements(env, secret, ndata, 0);
}

JNIEXPORT jint JNICALL SED_XXXGetSerialNumber(
    JNIEnv *env,
    jclass cls,
    jobject serial)
{
	if (serial == 0)
		return USK_ERR_JNI;
	
	jclass clazz = (*env)->GetObjectClass(env, serial);
	
	jmethodID mid = (*env)->GetMethodID (env, clazz, "append", "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
	if(mid == 0)
		return USK_ERR_JNI;
	char sn[128];
	//ZeroMemory(sn, 128);
	int ret = SmartDongleGetSerialNumber(sn);
	jstring jsn = (*env)->NewStringUTF(env, sn);
	(*env)->CallObjectMethod(env, serial, mid, jsn);
	return (jint)ret;
}
#endif
