/*
    smartdongle.h: SmartDongle libusb driver.

    MicroWorks, Inc.
    2808 North Cole Road
    Boise, ID 83704

    www.smartdongle.com


    ***** BEGIN LICENSE BLOCK *****

    Version: MPL 1.1/GPL 2.0/LGPL 2.1

    The contents of this file are subject to the Mozilla Public License
    Version 1.1 (the "License"); you may not use this file except in
    compliance with the License. You may obtain a copy of the License at
    http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS IS" basis,
    WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
    for the specific language governing rights and limitations under the
    License.

    The Original Code is MicroWorks, Inc. code.

    The Initial Developer of the Original Code is MicroWorks, Inc. Portions
    created by MicroWorks, Inc. are Copyright (C) 2008
    MicroWorks, Inc. All Rights Reserved.

    Contributor(s):

    Alternatively, the contents of this file may be used under the terms of
    either the GNU General Public License Version 2 or later (the "GPL"), or
    the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
    in which case the provisions of the GPL or the LGPL are applicable instead
    of those above. If you wish to allow use of your version of this file only
    under the terms of either the GPL or the LGPL, and not to allow others to
    use your version of this file under the terms of the MPL, indicate your
    decision by deleting the provisions above and replace them with the notice
    and other provisions required by the GPL or the LGPL. If you do not delete
    the provisions above, a recipient may use your version of this file under
    the terms of any one of the MPL, the GPL or the LGPL.

    ***** END LICENSE BLOCK *****
*/


#ifndef SMARTDONGLE_H
#define SMARTDONGLE_H

#include "usb.h"

#if defined(__cplusplus) || defined(__cplusplus__)
extern "C" {
#endif

/*
    Given the correct keys, P1 and P2, will place SmartDongle into
    enabled state, read data from SmartDongle user space into buffer,
    then disable SmartDongle.

    P1      Pointer to key #1
    P2      Pointer to key #2
    address Address into SmartDongle user memory. (0 thru 32431)
    buffer  Pointer to user allocated data buffer.
    size    Requested number of bytes to read from SmartDongle.
*/
int SmartDongleRead(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned short address,
    unsigned char *buffer,
    int size);
	

unsigned long long SmartDongleGetKeylessChallenge();

int SmartDongleSendKeylessResponse(unsigned long long response);

int SmartDongleKeylessWrite(
	unsigned short address,
	unsigned char *buffer,
	int size);
	
/*
 Read data from SmartDongle user space into buffer,
 then disable SmartDongle.
 
 address Address into SmartDongle user memory. (0 thru 32431)
 buffer  Pointer to user allocated data buffer.
 size    Requested number of bytes to read from SmartDongle.
 */
int SmartDongleKeylessRead(
	unsigned short address,
	unsigned char *buffer,
	int size);

/*
    Given the correct keys, P1 and P2, will place SmartDongle into
    enabled state, write data from buffer into SmartDongle user space,
    then disable SmartDongle.

    P1      Pointer to key #1
    P2      Pointer to key #2
    address Address into SmartDongle user memory. (0 thru 32431)
    buffer  Pointer to user allocated data buffer.
    size    Requested number of bytes to read from SmartDongle.
*/
int SmartDongleWrite(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned short address,
    unsigned char *buffer,
    int size);

/*
    LED Control.
*/
int SmartDongleLedGreen(void);
int SmartDongleLedRed(void);
int SmartDongleLedOff(void);

/*
    User EPROM memory encryption control for SmartDongle version 1.42 and above.

    Modes:

    0 - User EPROM memory space protected by strong (AES) encryption
    1 - User EPROM memory space is unencrypted
    2 - User EPROM memory space protected by fast encryption
*/
int SmartDongleWriteCryptMode(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned char mode);

/*
    Read current encryption mode.
*/
int SmartDongleReadCryptMode(unsigned char *mode);

/*
    Set current encryption mode. SmartDongle Version 1.70 and above
*/
int SmartDongleSetCryptMode(
    unsigned long long *P1,
    unsigned long long *P2,
    unsigned char *mode);

/*
    Returns SmartDongle Eeprom size
*/
int SmartDongleGetEepromSize(
    unsigned long long *P1,
    unsigned long long *P2);

/*
    For SmartDongle >=v1.50 using >= v2.3 driver only.
    This will configure the SmartDongle for use with
    the 2.3 driver only.

    Caution: This configuration is permanent. SmartDongle
              must be sent back to the manufacturer to be
              reset to use the older drivers (<= v2.2.0).
*/
int SmartDongleSetStrictOnly(void);

/*
    SmartDongle reboot.
*/
int SmartDongleReset(void);

/*
    Get SmartDongle serial number.
    User must allocate a serNo buffer of at least SMARTDNGL_SN_LENGTH + 1
*/
int SmartDongleGetSerialNumber (char *serNo);

/*
    Return error string corresponding to error code.
    User must allocate string size of SMARTDNGL_ERROR_STRING_LENGTH_MAX
*/
void SmartDongleGetErrorString(char *errorString, int error);

/*
    Maximum string length returned from GetErrorString()
*/
#define SMARTDNGL_ERROR_STRING_LENGTH_MAX 128

/*
    SmartDongle's EEPROM page size.
*/
#define SMARTDNGL_EPROMPGSZ 0x10

/*
    For serial number string descriptor.
*/
#define SMARTDNGL_SN_MAXLEN 32
#define SMARTDNGL_SN_LENGTH 12
#define SMARTDNGL_SN_INDEX  0x03

/*
    Three on-chip SmartDongle encrpytion modes to choose from.
    This provide protection for SmartDongle on-board memory only.
*/
#define SMARTDNGL_OPTION_ON_CHIP_EPROM_PROTECT_STRONG  0x00
#define SMARTDNGL_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE 0x01
#define SMARTDNGL_OPTION_ON_CHIP_EPROM_PROTECT_FAST    0x02

/*
    Error codes
*/
#define SMARTDNGL_ERR_OK               0

/* System memory allocation error */
#define SMARTDNGL_ERR_MEM              1

/* Address offset or buffer size exceeds EPROM memory */
#define SMARTDNGL_ERR_EPROM_SIZE       2

/* Timeout waiting for SmartDongle State change */
#define SMARTDNGL_ERR_SM_TIMEOUT       3

/* Timeout waiting for TWC to go to "inactive" state */
#define SMARTDNGL_ERR_TWC_TIMEOUT      4

/* Timeout waiting for Task Loop to finish */
#define SMARTDNGL_ERR_BSY_TIMEOUT      5

/* Timeout Waiting For Read Flag(s) to indicate either Page Ready or error. */
#define SMARTDNGL_ERR_EPROM_RD_TIMEOUT 6

/* Host failed to validate against SmartDongle keys */
#define SMARTDNGL_ERR_VALIDATE         7

/* Dongle failed to validate against user keys */
#define SMARTDNGL_ERR_USK_VALIDATE     35

/* Insufficient buffer size for command */
#define SMARTDNGL_ERR_BUFSIZE          8

/* Data format Error */
#define SMARTDNGL_ERR_DATA_FORMAT      36

/* One or more arguments are not correct */
#define SMARTDNGL_ERR_BAD_ARG          37

/* No SmartDongles found on host  */
#define SMARTDNGL_ERR_NODEV            9

/* For USB Vendor Device Request control failure. */
#define SMARTDNGL_ERR_CHALLENGE        10
#define SMARTDNGL_ERR_REPLY            11
#define SMARTDNGL_ERR_REQ_CHLG         12
#define SMARTDNGL_ERR_REQ_PAGE         13
#define SMARTDNGL_ERR_RD_STATE         14
#define SMARTDNGL_ERR_RD_STATUS        15
#define SMARTDNGL_ERR_RD_BSY           16
#define SMARTDNGL_ERR_RD_TWC           17
#define SMARTDNGL_ERR_RD_PGFLGS        18
#define SMARTDNGL_ERR_RD_EPROM         19
#define SMARTDNGL_ERR_WR_EPROM         20
#define SMARTDNGL_ERR_RESET_SM         21
#define SMARTDNGL_ERR_RESET_STATE      22
#define SMARTDNGL_ERR_LOCK             23
#define SMARTDNGL_ERR_SOFT_RST         24
#define SMARTDNGL_ERR_LED              25
#define SMARTDNGL_ERR_RD_MODE          26
#define SMARTDNGL_ERR_WR_MODE          27
#define SMARTDNGL_ERR_SET_STRICT       38
#define SMARTDNGL_ERR_RD_OPTIONS       39

#define SMARTDNGL_ERR_OPEN             32
#define SMARTDNGL_ERR_GET_SN           33
#define SMARTDNGL_ERR_DEV_MAX          34

/*
    Error attempting to use extended device functionality with old SmartDongle.
*/
#define SMARTDNGL_ERR_REQUIRES_1_50    40
#define SMARTDNGL_ERR_REQUIRES_1_70    49

/*
    The libusb command usb_open() failed.
*/
#define SMARTDNGL_ERR_USB_OPEN         41


/*
    The libusb command usb_claim_interface() failed.

    Error result from:
    1) The user has insufficient privileges to read files under /proc/bus/usb
    2) SmartDongle interface  may have been claimed by another process or thread.
    3) SmartDongle been unplugged.
*/
#define SMARTDNGL_ERR_USB_CLAIM_INTERFACE 42

/*
    The libusb_control_msg() failed.
*/
#define SMARTDNGL_ERR_USB_CONTROL_MSG  48

#define SMARTDNGL_ERR_UNKNOWN          255


#if defined(__cplusplus) || defined(__cplusplus__)
}
#endif

#endif  /* SMARTDONGLE_H */
